# -*- coding: utf-8 -*-
"""
Base classes and functions for region-specific services
"""

import re

from threading import Event, Thread
# Python2 vs Python3
try:
    from Queue import Queue
except ImportError:
    from queue import Queue

from opinel.utils.aws import build_region_list, connect_service, handle_truncated_response
from opinel.utils.console import printException, printInfo

from AWSScout2.utils import format_service_name
from AWSScout2.configs.base import GlobalConfig
from AWSScout2.output.console import FetchStatusLogger





########################################
# Globals
########################################

api_clients = dict()

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_caps_re = re.compile('([a-z0-9])([A-Z])')


########################################
# RegionalServiceConfig
########################################

class RegionalServiceConfig(object):
    """
    Single service configuration for non-global services

    :ivar regions:                      Dictionary of regions
    :ivar service:                      Name of the service
    """

    def __init__(self):
        self.regions = {}
        self.service = type(self).__name__.replace('Config', '').lower() # TODO: use regex with EOS instead of plain replace

    def init_region_config(self, region):
        """
        Initialize the region's configuration

        :param region:                  Name of the region
        """
        self.regions[region] = self.region_config_class()


    def fetch_all(self, credentials, regions = [], partition_name = 'aws', targets = None):
        """
        Fetch all the SNS configuration supported by Scout2

        :param credentials:             F
        :param service:                 Name of the service
        :param regions:                 Name of regions to fetch data from
        :param partition_name:          AWS partition to connect to
        :param targets:                 Type of resources to be fetched; defaults to all.

        """
        # Initialize targets
        if not targets:
            targets = type(self).targets
        printInfo('Fetching %s config...' % format_service_name(self.service))
        self.fetchstatuslogger = FetchStatusLogger(targets, True)
        api_service = 'ec2' if self.service.lower() == 'vpc' else self.service.lower()
        # Init regions
        regions = build_region_list(api_service, regions, partition_name) # TODO: move this code within this class
        self.fetchstatuslogger.counts['regions']['discovered'] = len(regions)
        # Threading to fetch & parse resources (queue consumer)
        q = self._init_threading(self._fetch_target, {}, 20)
        # Threading to list resources (queue feeder)
        qr = self._init_threading(self._fetch_region, {'api_service': api_service, 'credentials': credentials, 'q': q, 'targets': targets}, 10)
        # Go
        for region in regions:
            qr.put(region)
        # Join
        qr.join()
        q.join()
        # Show completion and force newline
        self.fetchstatuslogger.show(True)

    def _init_threading(self, function, params={}, num_threads=10):
            # Init queue and threads
            q = Queue(maxsize=0) # TODO: find something appropriate
            if not num_threads:
                num_threads = len(targets)
            for i in range(num_threads):
                worker = Thread(target=function, args=(q, params))
                worker.setDaemon(True)
                worker.start()
            return q

    def _fetch_region(self, q, params):
        global api_clients
        try:
            while True:
                try:
                    region = q.get()
                    self.init_region_config(region)
                    api_client = connect_service(params['api_service'], params['credentials'], region)
                    api_clients[region] = api_client
                    self.regions[region].fetch_all(api_client, self.fetchstatuslogger, params['q'], params['targets'])
                    self.fetchstatuslogger.counts['regions']['fetched'] += 1
                except Exception as e:
                    printException(e)
                finally:
                    q.task_done()
        except Exception as e:
            printException(e)
            pass

    def _fetch_target(self, q, params):
        try:
            while True:
                try:
                    method, region, target = q.get()
                    method(params, region, target)
                    target = method.__name__.replace('parse_', '') + 's'
                    self.fetchstatuslogger.counts[target]['fetched'] += 1
                    self.fetchstatuslogger.show()
                except Exception as e:
                    printException(e)
                finally:
                    q.task_done()
        except Exception as e:
            printException(e)
            pass

    def finalize(self):
        for t in self.fetchstatuslogger.counts:
            setattr(self, '%s_count' % t, self.fetchstatuslogger.counts[t]['fetched'])
        delattr(self, 'fetchstatuslogger')
        for r in self.regions:
            if hasattr(self.regions[r], 'fetchstatuslogger'):
                delattr(self.regions[r], 'fetchstatuslogger')



########################################
# RegionConfig
########################################

class RegionConfig(GlobalConfig):
    """
    Base class for ...
    """

    def __init__(self, region_name):
        self.region = region_name

    def fetch_all(self, api_client, fetchstatuslogger, q, targets):
        self.fetchstatuslogger = fetchstatuslogger
        if targets != None:
            # Ensure targets is a tuple
            if type(targets) != list and type(targets) != tuple:
                targets = tuple(targets,)
            elif type(targets) != tuple:
                targets = tuple(targets)
#        else:
#            targets = tuple(['%s' % method.replace('fetch_','').title() for method in methods])
        for target in targets:
            self._fetch_targets(api_client, q, target, {})


    def _fetch_targets(self, api_client, q, target, list_params):
        # Handle & format the target type
        target_type, response_attribute, list_method_name, ignore_list_error = target
        list_method = getattr(api_client, list_method_name)
        try:
            targets = handle_truncated_response(list_method, list_params, [response_attribute])[response_attribute]
        except Exception as e:
            if not ignore_list_error:
                printException(e)
            targets = []
        setattr(self, '%s_count' % target_type, len(targets))
        self.fetchstatuslogger.counts[target_type]['discovered'] += len(targets)
        region = api_client._client_config.region_name
        # Queue resources
        for target in targets:
            callback = getattr(self, 'parse_%s' % target_type[0:-1])
            if q:
                # Add to the queue
                q.put((callback, region, target))



# -*- coding: utf-8 -*-
"""
CloudWatch-related classes and functions
"""

from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig



########################################
# CloudWatchRegionConfig
########################################

class CloudWatchRegionConfig(RegionConfig):
    """
    CloudWatch configuration for a single AWS region

    :ivar trails:                       Dictionary of trails [name]
    :ivar trails_count:                 Number of trails in the region
    """

    def __init__(self):
        self.alarms = {}
        self.alarms_count = 0
        self.metrics = {}
        self.metrics_count = 0


    def parse_alarm(self, global_params, region, alarm):
        """
        Parse a single CloudWatch trail

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param alarm:                   Alarm
        """
        alarm['arn'] = alarm.pop('AlarmArn')
        alarm['name'] = alarm.pop('AlarmName')
        # Drop some data
        for k in ['AlarmConfigurationUpdatedTimestamp', 'StateReason', 'StateReasonData', 'StateUpdatedTimestamp']:
            foo = alarm.pop(k) if k in alarm else None
        alarm_id = self.get_non_aws_id(alarm['arn'])
        self.alarms[alarm_id] = alarm



########################################
# CloudWatchConfig
########################################

class CloudWatchConfig(RegionalServiceConfig):
    """
    CloudWatch configuration for all AWS regions

    :cvar targets:                      Tuple with all CloudWatch resource names that may be fetched
    :cvar region_config_class:          Class to be used when initiating the service's configuration in a new region
    """
    targets = (
        ('alarms', 'MetricAlarms', 'describe_alarms', False),
    )
    region_config_class = CloudWatchRegionConfig



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import json
import os

from opinel.utils.console import printException

from AWSScout2.configs.services import ServicesConfig

class Scout2Config(object):
    """
    Root object that holds all of the necessary AWS resources and Scout2
    configuration items.

    :account_id         AWS account ID
    :last_run           Information about the last run
    :metadata           Metadata used to generate the HTML report
    :ruleset            Ruleset used to perform the analysis
    :services           AWS configuration sorted by service
    """

    def __init__(self, profile, report_dir = None, timestamp = None, services= [], skipped_services = []):
        self.account_id = None
        self.last_run = None
        self.__load_metadata()
        self.services = ServicesConfig()
        supported_services = vars(self.services).keys()
        self.service_list = self.__build_services_list(supported_services, services, skipped_services)


    def fetch(self, credentials, regions = [], skipped_regions = [], partition_name = 'aws'):
        """

        :param credentials:
        :param services:
        :param skipped_services:
        :param regions:
        :param skipped_regions:
        :param partition_name:
        :return:
        """
        # TODO: determine partition name based on regions and warn if multiple partitions...
        self.services.fetch(credentials, self.service_list, regions, partition_name)


    def __load_metadata(self):
        # Load metadata
        scout2_configs_data_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
        with open(os.path.join(scout2_configs_data_path, 'metadata.json'), 'rt') as f:
            self.metadata = json.load(f)


    def __build_services_list(self, supported_services, services, skipped_services):
        return [s for s in supported_services if (services == [] or s in services) and s not in skipped_services]


    def update_metadata(self):
        service_map = {}
        for service_group in self.metadata:
            for service in self.metadata[service_group]:
                if service not in self.service_list:
                    continue
                if 'resources' not in self.metadata[service_group][service]:
                    continue
                service_map[service] = service_group
                for resource in self.metadata[service_group][service]['resources']:
                    # full_path = path if needed
                    if not 'full_path' in self.metadata[service_group][service]['resources'][resource]:
                        self.metadata[service_group][service]['resources'][resource]['full_path'] = self.metadata[service_group][service]['resources'][resource]['path']
                    # Script is the full path minus "id" (TODO: change that)
                    if not 'script' in self.metadata[service_group][service]['resources'][resource]:
                        self.metadata[service_group][service]['resources'][resource]['script'] = '.'.join([x for x in self.metadata[service_group][service]['resources'][resource]['full_path'].split('.') if x != 'id'])
                    # Update counts
                    count = '%s_count' % resource
                    service_config = getattr(self.services, service)
                    if service_config and resource != 'regions':
                      if hasattr(service_config, 'regions'):
                        self.metadata[service_group][service]['resources'][resource]['count'] = 0
                        for region in service_config.regions:
                            if hasattr(service_config.regions[region], count):
                                self.metadata[service_group][service]['resources'][resource]['count'] += getattr(service_config.regions[region], count)
                      else:
                          try:
                              self.metadata[service_group][service]['resources'][resource]['count'] = getattr(service_config, count)
                          except Exception as e:
                              printException(e)
                              print(vars(service_config))



# -*- coding: utf-8 -*-
"""
EC2-related classes and functions
"""

# TODO: move a lot of this to VPCconfig, and use some sort of filter to only list SGs in EC2 classic

from opinel.utils.aws import get_name
from opinel.utils.console import printException, printInfo
from opinel.utils.fs import load_data
from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig
from AWSScout2.utils import get_keys, ec2_classic


########################################
# Globals
########################################

icmp_message_types_dict = load_data('icmp_message_types.json', 'icmp_message_types')
protocols_dict = load_data('protocols.json', 'protocols')



########################################
# EC2RegionConfig
########################################

class EC2RegionConfig(RegionConfig):
    """
    EC2 configuration for a single AWS region

    :ivar vpcs:                         Dictionary of VPCs [id]
    :ivar instances_count:              Number of instances in the region
    :ivar parameter_groups:             Dictionary of parameter groups [id]
    :ivar parameter_groups_count:       Number of parameter groups in the region
    :ivar security_groups:              Dictionary of security groups [id]
    :ivar security_groups_count:        Number of security groups in the region
    """

    def __init__(self):
        self.vpcs = {}
        self.elastic_ips = {}
        self.elastic_ips_count = 0
        self.flow_logs_count = 0
        self.instances_count = 0
        self.network_acls_count = 0
        self.security_groups_count = 0
        self.subnets = {}               # This is a temporary artifact that is removed in the finalize() calls
        self.subnets_count = 0


    def parse_elastic_ip(self, global_params, region, eip):
        """

        :param global_params:
        :param region:
        :param eip:
        :return:
        """
        self.elastic_ips[eip['PublicIp']] = eip


    def parse_instance(self, global_params, region, reservation):
        """
        Parse a single EC2 instance

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param instance:                 Cluster
        """
        for i in reservation['Instances']:
            instance = {}
            vpc_id = i['VpcId'] if 'VpcId' in i and i['VpcId'] else ec2_classic
            manage_dictionary(self.vpcs, vpc_id, EC2VPCConfig())
            instance['reservation_id'] = reservation['ReservationId']
            instance['id'] = i['InstanceId']
            get_name(i, instance, 'InstanceId')
            get_keys(i, instance, ['KeyName', 'LaunchTime', 'InstanceType', 'State', 'IamInstanceProfile'])
            # Network interfaces & security groups
            manage_dictionary(instance, 'network_interfaces', {})
            for eni in i['NetworkInterfaces']:
                nic = {}
                get_keys(eni, nic, ['Association', 'Groups', 'PrivateIpAddresses'])
                instance['network_interfaces'][eni['NetworkInterfaceId']] = nic
            self.vpcs[vpc_id].instances[i['InstanceId']] = instance


    def parse_security_group(self, global_params, region, group):
        """
        Parse a single Redsfhit security group

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param security)_group:         Security group
        """
        vpc_id = group['VpcId'] if 'VpcId' in group and group['VpcId'] else ec2_classic
        manage_dictionary(self.vpcs, vpc_id, EC2VPCConfig())
        security_group = {}
        security_group['name'] = group['GroupName']
        security_group['id'] = group['GroupId']
        security_group['description'] = group['Description']
        security_group['owner_id'] = group['OwnerId']
        security_group['rules'] = {'ingress': {}, 'egress': {}}
        security_group['rules']['ingress']['protocols'], security_group['rules']['ingress']['count'] = self.__parse_security_group_rules(group['IpPermissions'])
        security_group['rules']['egress']['protocols'], security_group['rules']['egress']['count'] = self.__parse_security_group_rules(group['IpPermissionsEgress'])
        self.vpcs[vpc_id].security_groups[group['GroupId']] = security_group


    def __parse_security_group_rules(self, rules):
        """

        :param self:
        :param rules:
        :return:
        """
        protocols = {}
        rules_count = 0
        for rule in rules:
            ip_protocol = rule['IpProtocol'].upper()
            if ip_protocol == '-1':
                ip_protocol = 'ALL'
            protocols = manage_dictionary(protocols, ip_protocol, {})
            protocols[ip_protocol] = manage_dictionary(protocols[ip_protocol], 'ports', {})
            # Save the port (single port or range)
            port_value = 'N/A'
            if 'FromPort' in rule and 'ToPort' in rule:
                if ip_protocol == 'ICMP':
                    # FromPort with ICMP is the type of message
                    port_value = icmp_message_types_dict[str(rule['FromPort'])]
                elif rule['FromPort'] == rule['ToPort']:
                    port_value = str(rule['FromPort'])
                else:
                    port_value = '%s-%s' % (rule['FromPort'], rule['ToPort'])
            manage_dictionary(protocols[ip_protocol]['ports'], port_value, {})
            # Save grants, values are either a CIDR or an EC2 security group
            for grant in rule['UserIdGroupPairs']:
                manage_dictionary(protocols[ip_protocol]['ports'][port_value], 'security_groups', [])
                protocols[ip_protocol]['ports'][port_value]['security_groups'].append(grant)
                rules_count = rules_count + 1
            for grant in rule['IpRanges']:
                manage_dictionary(protocols[ip_protocol]['ports'][port_value], 'cidrs', [])
                protocols[ip_protocol]['ports'][port_value]['cidrs'].append({'CIDR': grant['CidrIp']})
                rules_count = rules_count + 1
        return protocols, rules_count



########################################
# EC2Config
########################################

class EC2Config(RegionalServiceConfig):
    """
    EC2 configuration for all AWS regions

    :cvar targets:                      Tuple with all EC2 resource names that may be fetched
    :cvar config_class:                 Class to be used when initiating the service's configuration in a new region/VPC
    """
    targets = (
        ('security_groups', 'SecurityGroups', 'describe_security_groups', False),
        ('instances', 'Reservations', 'describe_instances', False)
    )
    region_config_class = EC2RegionConfig



########################################
# EC2VPCConfig
########################################

class EC2VPCConfig(object):
    """
    EC2 configuration for a single VPC

    :ivar flow_logs:                    Dictionary of flow logs [id]
    :ivar instances:                    Dictionary of instances [id]
    """

    def __init__(self, name = None):
        self.name = name
        self.flow_logs = {}
        self.instances = {}
        self.network_acls = {}
        self.route_tables = {}
        self.security_groups = {}
        self.subnets = {}



########################################
##### EC2 analysis functions
########################################

def analyze_ec2_config(ec2_info, aws_account_id, force_write):
    try:
        printInfo('Analyzing EC2 config... ', newLine = False)
        # Tweaks
        link_elastic_ips(ec2_info)
        add_security_group_name_to_ec2_grants(ec2_info, aws_account_id)
        # Custom EC2 analysis
#        check_for_elastic_ip(ec2_info)
        list_network_attack_surface(ec2_info, 'attack_surface', 'PublicIpAddress')
        # TODO: make this optional, commented out for now
        # list_network_attack_surface(ec2_info, 'private_attack_surface', 'PrivateIpAddress')
        printInfo('Success')
    except Exception as e:
        printInfo('Error')
        printException(e)

#
# Github issue #24: display the security group names in the list of grants (added here to have ligher JS code)
#
def add_security_group_name_to_ec2_grants(ec2_config, aws_account_id):
    go_to_and_do(ec2_config, None, ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports', 'security_groups'], [], add_security_group_name_to_ec2_grants_callback, {'AWSAccountId': aws_account_id})
#
# Callback
#
def add_security_group_name_to_ec2_grants_callback(ec2_config, current_config, path, current_path, ec2_grant, callback_args):
    sg_id = ec2_grant['GroupId']
    if sg_id in current_path:
        target = current_path[:(current_path.index(sg_id)+1)]
        ec2_grant['GroupName'] = get_attribute_at(ec2_config, target, 'name')
    elif ec2_grant['UserId'] == callback_args['AWSAccountId']:
        if 'VpcId' in ec2_grant:
            target = current_path[:(current_path.index('vpcs')+1)]
            target.append(ec2_grant['VpcId'])
            target.append('security_groups')
            target.append(sg_id)
        else:
            target = current_path[:(current_path.index('security_groups')+1)]
            target.append(sg_id)
        ec2_grant['GroupName'] = get_attribute_at(ec2_config, target, 'name')

#
# Check that the whitelisted EC2 IP addresses are not static IPs
#
def check_for_elastic_ip(ec2_info):
    # Build a list of all elatic IP in the account
    elastic_ips = []
    for region in ec2_info['regions']:
        if 'elastic_ips' in ec2_info['regions'][region]:
            for eip in ec2_info['regions'][region]['elastic_ips']:
                elastic_ips.append(eip)
    new_items = []
    new_macro_items = []
    for i, item in enumerate(ec2_info['violations']['non-elastic-ec2-public-ip-whitelisted'].items):
        ip = netaddr.IPNetwork(item)
        found = False
        for eip in elastic_ips:
            eip = netaddr.IPNetwork(eip)
            if ip in eip:
                found = True
                break
        if not found:
            new_items.append(ec2_info['violations']['non-elastic-ec2-public-ip-whitelisted'].items[i])
            new_macro_items.append(ec2_info['violations']['non-elastic-ec2-public-ip-whitelisted'].macro_items[i])
    ec2_info['violations']['non-elastic-ec2-public-ip-whitelisted'].items = new_items
    ec2_info['violations']['non-elastic-ec2-public-ip-whitelisted'].macro_items = new_macro_items


#
# Link EIP with instances (looks like this might be no longer needed with boto3)
#
def link_elastic_ips(ec2_config):
    return
    go_to_and_do(ec2_config, None, ['regions', 'elastic_ips'], None, link_elastic_ips_callback1, {})

def link_elastic_ips_callback1(ec2_config, current_config, path, current_path, elastic_ip, callback_args):
    if not 'id' in current_config:
        return
    instance_id = current_config['id']
    return
    go_to_and_do(ec2_config, None, ['regions', 'vpcs', 'instances'], None, link_elastic_ips_callback2, {'instance_id': instance_id, 'elastic_ip': elastic_ip})

def link_elastic_ips_callback2(ec2_config, current_config, path, current_path, instance_id, callback_args):
    if instance_id == callback_args['instance_id']:
        if not 'PublicIpAddress' in current_config:
            current_config['PublicIpAddress'] = callback_args['elastic_ip']
        elif current_config['PublicIpAddress'] != callback_args['elastic_ip']:
            printInfo('Warning: public IP address exists (%s) for an instance associated with an elastic IP (%s)' % (current_config['PublicIpAddress'], callback_args['elastic_ip']))
            # This can happen... fix it







#
# Once all the data has been fetched, iterate through instances and list them
# Could this be done when all the "used_by" values are set ??? TODO
#
def list_instances_in_security_groups(region_info):
    for vpc in region_info['vpcs']:
        if not 'instances' in region_info['vpcs'][vpc]:
            return
        for instance in region_info['vpcs'][vpc]['instances']:
            state = region_info['vpcs'][vpc]['instances'][instance]['State']['Name']
            for sg in region_info['vpcs'][vpc]['instances'][instance]['security_groups']:
                sg_id = sg['GroupId']
                manage_dictionary(region_info['vpcs'][vpc]['security_groups'][sg_id], 'instances', {})
                manage_dictionary(region_info['vpcs'][vpc]['security_groups'][sg_id]['instances'], state, [])
                region_info['vpcs'][vpc]['security_groups'][sg_id]['instances'][state].append(instance)




#
# Ensure name and ID are set
#
def manage_vpc(vpc_info, vpc_id):
    manage_dictionary(vpc_info, vpc_id, {})
    vpc_info[vpc_id]['id'] = vpc_id
    if not 'name' in vpc_info[vpc_id]:
        vpc_info[vpc_id]['name'] = vpc_id







# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import fnmatch
import json
import os
import re
import shutil

from opinel.utils.console import printDebug, printError, printException, printInfo
from opinel.utils.fs import read_ip_ranges
from opinel.utils.globals import manage_dictionary

from AWSScout2.rules import condition_operators
from AWSScout2.rules.utils import recurse

finding_levels = ['danger', 'warning']

# First search local under ./rules and ./rulesets
# Then search from package files
# Then error


re_ip_ranges_from_file = re.compile(r'_IP_RANGES_FROM_FILE_\((.*?),\s*(.*?)\)')
re_get_value_at = re.compile(r'_GET_VALUE_AT_\((.*?)\)')
re_list_value = re.compile(r'_LIST_\((.*?)\)')
aws_ip_ranges_filename = 'ip-ranges.json'
ip_ranges_from_args = 'ip-ranges-from-args'

FILTERS_DIR = 'filters'
FINDINGS_DIR = 'findings'
RULESETS_DIR = 'rulesets'
DEFAULT_RULESET = '%s/default.json' % RULESETS_DIR


class Ruleset(object):
    """
    TODO
    """

    def __init__(self, environment_name = 'default', filename = None, name = None, services = [], load_ruleset = True, load_rules = True, rule_type = 'findings', rules_dir = None, ip_ranges = []):
        self.rules_data_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
        self.environment_name = environment_name
        # Ruleset filename
        self.filename = self.find_file(filename)
        if not self.filename:
            self.search_ruleset(environment_name)
        self.name = os.path.basename(self.filename).replace('.json','') if not name else name
        # Load ruleset
        self.ruleset = {}
        if load_ruleset:
            self.load_ruleset()
        self.update_ruleset(rules_dir)
        # Rules or filters ?
        self.rule_type = rule_type
        # Load rules
        self.rules = {}
        if load_rules:
            aws_account_id = ''
            self.init_rules(services, ip_ranges, aws_account_id, False)


    def load_ruleset(self, quiet = False):
        if self.filename and os.path.exists(self.filename):
            try:
                with open(self.filename, 'rt') as f:
                    self.ruleset = json.load(f)
            except Exception as e:
                printException(e)
                printError('Error: ruleset file %s contains malformed JSON.' % self.filename)
                self.ruleset = {}
        else:
            self.ruleset = {}
            if not quiet:
                printError('Error: the file %s does not exist.' % self.filename)

    def update_ruleset(self, rules_dir):
        if rules_dir == None:
            return
        self.available_rules = {}
        parameterized_rules = []
        self.services = []
        for rule in self.ruleset['rules']:
            rule['filename'] = rule['filename'].replace('rules/', '')
            if not 'args' in rule:
                self.available_rules[rule['filename']] = rule
            else:
                parameterized_rules.append(rule)
        # Add default location
        rules_dir.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/findings'))
        for dir in rules_dir:
            rule_filenames = [f for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]
            for rule_filename in rule_filenames:
                self.services.append(rule_filename.split('-')[0].lower())
                printDebug('Loading %s' % rule_filename)
                with open('%s/%s' % (dir, rule_filename), 'rt') as f:
                    rule = json.load(f)
                    if not 'key' in rule and not 'arg_names' in rule:
                        # Non-parameterized rule, save it
                        if rule_filename in self.available_rules:
                            self.available_rules[rule_filename].update(rule)
                        else:
                            self.available_rules[rule_filename] = rule
                            self.available_rules[rule_filename]['enabled'] = False
                            if 'level' not in self.available_rules[rule_filename]:
                                self.available_rules[rule_filename]['level'] = 'danger'
                                self.available_rules[rule_filename]['filename'] = rule_filename
                    else:
                        # Parameterized rules, find all occurences and save N times
                        parameterized_rule_found = False
                        for prule in parameterized_rules:
                            if prule['filename'] == rule_filename:
                                parameterized_rule_found = True
                                for k in rule:
                                    prule[k] = set_argument_values(rule[k], prule['args'],
                                                                   convert=True) if k != 'conditions' else rule[k]
                                key = prule.pop('key') if 'key' in prule else prule['filename']
                                args = prule.pop('args')
                                if not 'arg_names' in prule:
                                    printError('No arg names key in %s' % rule_filename)
                                    continue
                                arg_names = prule.pop('arg_names')
                                if len(args) != len(arg_names):
                                    printError('Error: rule %s expects %d arguments but was provided %d.' % (
                                    rule_filename, len(arg_names), len(args)))
                                    continue
                                prule['args'] = []
                                for (arg_name, arg_value) in zip(arg_names, args):
                                    prule['args'].append({'arg_name': arg_name, 'arg_value': arg_value})
                                if 'level' not in prule:
                                    prule['level'] = 'danger'
                                    self.available_rules[key] = prule
                        if not parameterized_rule_found:
                            # Save once with no parameters
                            self.available_rules[rule_filename] = rule
                            self.available_rules[rule_filename]['enabled'] = False
                            if 'level' not in self.available_rules[rule_filename]:
                                self.available_rules[rule_filename]['level'] = 'danger'
                                self.available_rules[rule_filename]['filename'] = rule_filename
                            args = []
                            for a in rule['arg_names']:
                                args.append({'arg_name': a, 'arg_value': ''})
                                self.available_rules[rule_filename]['args'] = args
                            printDebug('Saving rule without parameter value: %s' % rule_filename)


    def html_generator(self, output_dir, force_write, debug):
        # Prepare the output directories
        prepare_html_output_dir(output_dir)
        # Create the JS include file
        printInfo('Preparing the HTML ruleset generator...')
        js_ruleset = {}
        js_ruleset['name'] = self.name
        js_ruleset['available_rules'] = self.available_rules
        js_ruleset['services'] = list(sorted(set(self.services)))
        save_config_to_file(self.environment_name, js_ruleset, 'ruleset', output_dir, force_write, debug)
        # Create the HTML generator
        html_generator = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../rules/data/ruleset-generator.html')
        dst_html_generator = os.path.join(output_dir, 'ruleset-generator.html')
        shutil.copyfile(html_generator, dst_html_generator)
        return dst_html_generator





    def search_ruleset(self, environment_name):
        ruleset_found = False
        if environment_name != 'default':
            for f in os.listdir(os.getcwd()):
                if fnmatch.fnmatch(f, '*.' + environment_name + '.json'):
                    ruleset_found = True
            if ruleset_found and prompt_4_yes_no("A ruleset whose name matches your environment name was found in %s. Would you like to use it instead of the default one" % f):
                self.filename = f
        if not ruleset_found:
            self.filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/rulesets/default.json')




    #
    # Initialize rules based on ruleset and services in scope
    #
    def init_rules(self, services, ip_ranges, aws_account_id, generator = False):
        # Load rules from JSON files
        for rule_metadata in self.ruleset['rules']:
            # Skip disabled rules
            if 'enabled' in rule_metadata and rule_metadata['enabled'] in ['false', 'False', False] and not generator:
                continue
            # Skip rules that apply to an out-of-scope service
            rule_details = self.load_json_rule(rule_metadata, ip_ranges, aws_account_id)
            if not rule_details:
                continue
            if 'enabled' in rule_metadata and rule_metadata['enabled']:
                rule_details['enabled'] = True
            skip_rule = True
            for service in services:
                if rule_details['path'].startswith(service):
                    skip_rule = False
#            if skip_rule:
#                continues
            #  Build the rules dictionary
            path = rule_details['path']
            if 'level' in rule_metadata:
                rule_details['level'] = rule_metadata['level']
            key = rule_details['key'] if 'key' in rule_details else rule_metadata['filename']
            # Set condition operator
            if not 'condition_operator' in rule_details:
                rule_details['condition_operator'] = 'and'
            # Save details for rule
            key = key.replace('.json', '').replace(' ', '')
            manage_dictionary(self.rules, path, {})
            self.rules[path][key] = rule_details



    #
    # Load rule from a JSON config file
    #
    def load_json_rule(self, rule_metadata, ip_ranges, aws_account_id):
        config = None
        config_file = rule_metadata['filename']
        config_args = rule_metadata['args'] if 'args' in rule_metadata else []
        # Determine the file path
        if not os.path.isfile(config_file):
            # Not a valid relative / absolute path, check locally under findings/ or filters/
            if not config_file.startswith('findings/') and not config_file.startswith('filters/'):
                config_file = '%s/%s' % (self.rule_type, config_file)
            if not os.path.isfile(config_file):
                config_file = os.path.join(self.rules_data_path, config_file)
        # Read the file
        try:
            #print('Reading %s' % config_file)
            with open(config_file, 'rt') as f:
                config = f.read()
            # Replace arguments
            for idx, argument in enumerate(config_args):
                config = config.replace('_ARG_'+str(idx)+'_', str(argument).strip())
            config = json.loads(config)
            config['filename'] = rule_metadata['filename']
            if 'args' in rule_metadata:
                config['args'] = rule_metadata['args']
            # Load lists from files
            for c1 in config['conditions']:
                if c1 in condition_operators:
                    continue
                if not type(c1[2]) == list and not type(c1[2]) == dict:
                    values = re_ip_ranges_from_file.match(c1[2])
                    if values:
                        filename = values.groups()[0]
                        conditions = json.loads(values.groups()[1])
                        if filename == aws_ip_ranges_filename:
                             c1[2] = read_ip_ranges(aws_ip_ranges_filename, False, conditions, True)
                        elif filename == ip_ranges_from_args:
                            c1[2] = []
                            for ip_range in ip_ranges:
                                c1[2] = c1[2] + read_ip_ranges(ip_range, True, conditions, True)
                    if c1[2] and aws_account_id:
                        if not type(c1[2]) == list:
                            c1[2] = c1[2].replace('_AWS_ACCOUNT_ID_', aws_account_id)
    
                    # Set lists
                    list_value = re_list_value.match(str(c1[2]))
                    if list_value:
                        values = []
                        for v in list_value.groups()[0].split(','):
                            values.append(v.strip())
                        c1[2] = values
        except Exception as e:
            printException(e)
            printError('Error: failed to read the rule from %s' % config_file)
        return config


    def analyze(self, aws_config):
        """

        :param aws_config:
        """
        printInfo('Analyzing AWS config...')
        # TODO: reset violations for all services in scope (maybe this can be done somewhere else (e.g. loading)
        for finding_path in self.rules:
            for rule in self.rules[finding_path]:
                printDebug('Processing %s rule[%s]: "%s"' % (finding_path.split('.')[0], self.rule_type[:-1], self.rules[finding_path][rule]['description']))
                path = finding_path.split('.')
                service = path[0]
                manage_dictionary(aws_config['services'][service], self.rule_type, {})
                aws_config['services'][service][self.rule_type][rule] = {}
                aws_config['services'][service][self.rule_type][rule]['description'] = self.rules[finding_path][rule]['description']
                aws_config['services'][service][self.rule_type][rule]['path'] = self.rules[finding_path][rule]['path']
                if self.rule_type == 'findings':
                    aws_config['services'][service][self.rule_type][rule]['level'] = self.rules[finding_path][rule]['level']
                if 'id_suffix' in self.rules[finding_path][rule]:
                    aws_config['services'][service][self.rule_type][rule]['id_suffix'] = self.rules[finding_path][rule]['id_suffix']
                if 'display_path' in self.rules[finding_path][rule]:
                    aws_config['services'][service][self.rule_type][rule]['display_path'] = self.rules[finding_path][rule]['display_path']
                try:
                    aws_config['services'][service][self.rule_type][rule]['items'] = recurse(aws_config['services'], aws_config['services'], path, [], self.rules[finding_path][rule], True)
                    aws_config['services'][service][self.rule_type][rule]['dashboard_name'] = self.rules[finding_path][rule]['dashboard_name'] if 'dashboard_name' in self.rules[finding_path][rule] else '??'
                    aws_config['services'][service][self.rule_type][rule]['checked_items'] = self.rules[finding_path][rule]['checked_items'] if 'checked_items' in self.rules[finding_path][rule] else 0
                    aws_config['services'][service][self.rule_type][rule]['flagged_items'] = len(aws_config['services'][service][self.rule_type][rule]['items'])
                    aws_config['services'][service][self.rule_type][rule]['service'] = service
                    aws_config['services'][service][self.rule_type][rule]['rationale'] = self.rules[finding_path][rule]['rationale'] if 'rationale' in self.rules[finding_path][rule] else 'N/A'
                except Exception as e:
                    printError('Failed to process rule defined in %s.json' % rule)
                    # Fallback if process rule failed to ensure report creation and data dump still happen
                    aws_config['services'][service][self.rule_type][rule]['checked_items'] = 0
                    aws_config['services'][service][self.rule_type][rule]['flagged_items'] = 0
                    printException(e)


    def find_file(self, filename, filetype = 'rulesets'):
        """

        :param filename:
        :param filetype:
        :return:
        """
        if filename and not os.path.isfile(filename):
            # Not a valid relative / absolute path, check Scout2's data under findings/ or filters/
            if not filename.startswith('findings/') and not filename.startswith('filters/'):
                filename = '%s/%s' % (filetype, filename)
            if not os.path.isfile(filename):
                filename = os.path.join(self.rules_data_path, filename)
            if not os.path.isfile(filename) and not filename.endswith('.json'):
                filename = self.find_file('%s.json' % filename, filetype)
        return filename


def set_argument_values(parameterized_input, target, convert = False):
    if convert:
        parameterized_input = json.dumps(parameterized_input)
    args = re.findall(r'(_ARG_(\w+)_)', parameterized_input)
    for arg in args:
        index = int(arg[1])
        parameterized_input = parameterized_input.replace(arg[0], target[index])
    return json.loads(parameterized_input) if convert else parameterized_input




# -*- coding: utf-8 -*-

import json

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients



########################################
# SQSRegionConfig
########################################

class SQSRegionConfig(RegionConfig):
    """
    SQS configuration for a single AWS region

    :ivar queues:                       Dictionary of queues [name]
    :ivar queues_count:                 Number of queues in the region
    """

    def __init__(self):
        self.queues = {}
        self.queues_count = 0


    def parse_queue(self, global_params, region, queue_url):
        """
        Parse a single queue and fetch additional attributes

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param queue_url:               URL of the AWS queue
        """
        queue = {'QueueUrl': queue_url}
        attributes = api_clients[region].get_queue_attributes(QueueUrl = queue_url, AttributeNames = ['CreatedTimestamp', 'Policy', 'QueueArn'])['Attributes']
        queue['arn'] = attributes.pop('QueueArn')
        for k in ['CreatedTimestamp']:
            queue[k] = attributes[k] if k in attributes else None
        if 'Policy' in attributes:
            queue['Policy'] = json.loads(attributes['Policy'])
        else:
            queue['Policy'] = {'Statement': []}

        queue['name'] = queue['arn'].split(':')[-1]
        self.queues[queue['name']] = queue



########################################
# SQSConfig
########################################

class SQSConfig(RegionalServiceConfig):
    """
    SQS configuration for all AWS regions

    :cvar targets:                      Tuple with all SQS resource names that may be fetched
    :cvar region_config_class:          Class to be used when initiating the service's configuration in a new region
    """
    targets = (
        ('queues', 'QueueUrls', 'list_queues', False),
    )
    region_config_class = SQSRegionConfig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import sys

try:
    from opinel.utils.globals import check_requirements
    from opinel.utils.console import configPrintException, printInfo
except Exception as e:
    print('Error: Scout2 depends on the opinel package. Install all the requirements with the following command:')
    print('  $ pip install -r requirements.txt')
    print(e)
    sys.exit(42)

from AWSScout2 import AWSCONFIG
from AWSScout2.cli_parser import ListallArgumentParser
from AWSScout2.output.console import format_listall_output, generate_listall_output
from AWSScout2.output.html import Scout2Report
from AWSScout2.rules.ruleset import Ruleset
from AWSScout2.rules.utils import recurse


########################################
##### Main
########################################

def main():

    # Parse arguments
    parser = ListallArgumentParser()
    args = parser.parse_args()

    # Configure the debug level
    configPrintException(args.debug)

    # Check version of opinel
    if not check_requirements(os.path.realpath(__file__)):
        return 42

    # Support multiple environments
    for profile_name in args.profile:

        # Load the config
        report = Scout2Report(profile_name, args.report_dir, args.timestamp)
        aws_config = report.jsrw.load_from_file(AWSCONFIG)
        services = aws_config['service_list']

        # Create a ruleset with only whatever rules were specified...
        if args.config:
            ruleset = Ruleset(filename = 'sample', load_rules = False)
            ruleset.ruleset['rules'][0]['filename'] = args.config
            ruleset.init_rules(services, args.ip_ranges, '', False) # aws_config['aws_account_id, False)
            # Need to set the arguments values args.config_args
        else:
            # TODO:
            #args = args
            #config = {}
            #config['conditions'] = args.conditions if hasattr(args, 'conditions') else []
            #config['mapping'] = args.mapping if hasattr(args, 'mapping') else []
            pass

        # Get single rule... TODO: clean
        tmp = ruleset.rules.pop(ruleset.rules.keys()[0])
        rule = tmp.pop(tmp.keys()[0])

        # Set the keys to output
        if len(args.keys):
            # 1. Explicitly provided on the CLI
            rule['keys'] = args.keys
        elif len(args.keys_file):
            # 2. Explicitly provided files that contain the list of keys
            rule['keys'] = []
            for filename in args.keys_file:
                with open(filename, 'rt') as f:
                    rule['keys'] += json.load(f)['keys']
        else:
            try:
                # 3. Load default set of keys based on path
                target_path = config['display_path'] if 'display_path' in config else config['path']
                with open('listall-configs/%s.json' % target_path) as f:
                    rule['keys'] = json.load(f)['keys']
            except:
                # 4. Print the object name
                rule['keys'] = ['name']

        # Recursion
        if len(args.path):
            rule['path'] = args.path[0]
        target_path = rule['path'].split('.')
        current_path = []
        resources = recurse(aws_config['services'], aws_config['services'], target_path, current_path, rule)

        # Prepare the output format
        (lines, template) = format_listall_output(args.format_file, 'foo', args.format, rule)

        # Print the output
        printInfo(generate_listall_output(lines, resources, aws_config, template, []))


#
# Load rule from a JSON config file
#
def load_config_from_json(rule_metadata, ip_ranges, aws_account_id, rule_type = 'rules'):
    config = None
    config_file = rule_metadata['filename']
    if not config_file.startswith('rules/') and not config_file.startswith('filters/'):
        config_file = '%s/%s' % (rule_type, config_file)
    config_args = rule_metadata['args'] if 'args' in rule_metadata else []
    try:
        with open(config_file, 'rt') as f:
            config = f.read()
        # Replace arguments
        for idx, argument in enumerate(config_args):
            config = config.replace('_ARG_'+str(idx)+'_', str(argument).strip())
        config = json.loads(config)
        config['filename'] = rule_metadata['filename']
        if 'args' in rule_metadata:
            config['args'] = rule_metadata['args']
        # Load lists from files
        for c1 in config['conditions']:
            if c1 in condition_operators:
                continue
            if not type(c1[2]) == list and not type(c1[2]) == dict:
                values = re_ip_ranges_from_file.match(c1[2])
                if values:
                    filename = values.groups()[0]
                    conditions = json.loads(values.groups()[1])
                    if filename == aws_ip_ranges_filename:
                         c1[2] = read_ip_ranges(aws_ip_ranges_filename, False, conditions, True)
                    elif filename == ip_ranges_from_args:
                        c1[2] = []
                        for ip_range in ip_ranges:
                            c1[2] = c1[2] + read_ip_ranges(ip_range, True, conditions, True)
                if c1[2] and aws_account_id:
                    if not type(c1[2]) == list:
                        c1[2] = c1[2].replace('_AWS_ACCOUNT_ID_', aws_account_id)

                # Set lists
                list_value = re_list_value.match(str(c1[2]))
                if list_value:
                    values = []
                    for v in list_value.groups()[0].split(','):
                        values.append(v.strip())
                    c1[2] = values
    except Exception as e:
        printException(e)
        printError('Error: failed to read the rule from %s' % config_file)
    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import os
import sys
import webbrowser

try:
    from opinel.utils.console import configPrintException, printInfo
    from opinel.utils.credentials import read_creds
    from opinel.utils.globals import check_requirements
except Exception as e:
    print('Error: Scout2 depends on the opinel package. Install all the requirements with the following command:')
    print('  $ pip install -r requirements.txt')
    print(e)
    sys.exit(42)

from AWSScout2 import AWSCONFIG, __version__ as scout2_version
from AWSScout2.cli_parser import Scout2ArgumentParser
from AWSScout2.configs.scout2 import  Scout2Config
from AWSScout2.output.html import Scout2Report
from AWSScout2.rules.exceptions import process_exceptions
from AWSScout2.rules.ruleset import Ruleset
from AWSScout2.rules.preprocessing import preprocessing
from AWSScout2.rules.postprocessing import postprocessing


########################################
##### Main
########################################

def main():

    # Parse arguments
    parser = Scout2ArgumentParser()
    args = parser.parse_args()

    # Configure the debug level
    configPrintException(args.debug)

    # Check version of opinel
    if not check_requirements(os.path.realpath(__file__)):
        return 42

    # Set the profile name
    profile_name = args.profile[0]

    # Search for AWS credentials
    if not args.fetch_local:
        credentials = read_creds(args.profile[0], args.csv_credentials, args.mfa_serial, args.mfa_code)
        if credentials['AccessKeyId'] is None:
            return 42

    # Create a new Scout2 config
    report = Scout2Report(profile_name, args.report_dir, args.timestamp)
    aws_config = Scout2Config(profile_name, args.report_dir, args.timestamp, args.services, args.skipped_services)

    if not args.fetch_local:

        # Fetch data from AWS APIs if not running a local analysis
        try:
            aws_config.fetch(credentials, regions=args.regions, partition_name=args.partition_name)
        except KeyboardInterrupt:
            printInfo('\nCancelled by user')
            return 130
        aws_config = report.jsrw.to_dict(aws_config)

        # Update means we reload the whole config and overwrite part of it
        if args.update == True:
            new_aws_config = copy.deepcopy(aws_config)
            aws_config = report.jsrw.load_from_file(AWSCONFIG)
            for service in new_aws_config['service_list']:
                # Per service only for now, may add per region & per VPC later...
                aws_config['services'][service] = new_aws_config['services'][service]

    else:

        # Reload to flatten everything into a python dictionary
        aws_config = report.jsrw.load_from_file(AWSCONFIG)

    # Pre processing
    preprocessing(aws_config, args.ip_ranges, args.ip_ranges_name_key)

    # Analyze config
    ruleset = Ruleset(profile_name, filename = args.ruleset, ip_ranges = args.ip_ranges)
    ruleset.analyze(aws_config)

    # Create display filters
    filters = Ruleset(filename = 'filters.json', rule_type = 'filters')
    filters.analyze(aws_config)

    # Handle exceptions
    process_exceptions(aws_config, args.exceptions[0])

    # Finalize
    postprocessing(aws_config, report.current_time, ruleset)

    # Save config and create HTML report
    html_report_path = report.save(aws_config, {}, args.force_write, args.debug)

    # Open the report by default
    if not args.no_browser:
        printInfo('Opening the HTML report...')
        url = 'file://%s' % os.path.abspath(html_report_path)
        webbrowser.open(url, new=2)




# -*- coding: utf-8 -*-

from opinel.utils.aws import handle_truncated_response
from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients
from AWSScout2.utils import ec2_classic



########################################
# RDSRegionConfig
########################################

class RDSRegionConfig(RegionConfig):
    """
    RDS configuration for a single AWS region

    :ivar vpcs:                         Dictionary of VPCs [id]
    :ivar instances_count:              Number of instances in the region
    :ivar parameter_groups:             Dictionary of parameter groups [id]
    :ivar parameter_groups_count:       Number of parameter groups in the region
    :ivar security_groups_count:        Number of security groups in the region
    """

    def __init__(self):
        self.vpcs = {}
        self.instances_count = 0
        self.parameter_groups = {}
        self.parameter_groups_count = 0
        self.security_groups_count = 0


    def parse_instance(self, global_params, region, dbi):
        """
        Parse a single RDS instance

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param instance:                Instance
        """
        vpc_id = dbi['DBSubnetGroup']['VpcId'] if 'DBSubnetGroup' in dbi and 'VpcId' in dbi['DBSubnetGroup'] and dbi['DBSubnetGroup']['VpcId'] else ec2_classic
        instance = {}
        instance['name'] = dbi.pop('DBInstanceIdentifier')
        for key in ['InstanceCreateTime', 'Engine', 'DBInstanceStatus', 'AutoMinorVersionUpgrade',
                    'DBInstanceClass', 'MultiAZ', 'Endpoint', 'BackupRetentionPeriod', 'PubliclyAccessible',
                    'StorageEncrypted', 'VpcSecurityGroups', 'DBSecurityGroups', 'DBParameterGroups',
                    'EnhancedMonitoringResourceArn', 'StorageEncrypted']:
                    # parameter_groups , security_groups, vpc_security_groups
            instance[key] = dbi[key] if key in dbi else None
        # If part of a cluster, multi AZ information is only available via cluster information
        if 'DBClusterIdentifier' in dbi:
            api_client = api_clients[region]
            cluster = api_client.describe_db_clusters(DBClusterIdentifier = dbi['DBClusterIdentifier'])['DBClusters'][0]
            instance['MultiAZ'] = cluster['MultiAZ']
        # Save
        manage_dictionary(self.vpcs, vpc_id,RDSVPCConfig())
        self.vpcs[vpc_id].instances[instance['name']] = instance


    def parse_snapshot(self, global_params, region, dbs):
        """

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param dbs:                     Snapshot
        :return:
        """
        vpc_id = dbs['VpcId'] if 'VpcId' in dbs else ec2_classic
        snapshot_id = dbs.pop('DBSnapshotIdentifier')
        snapshot = {'arn': dbs.pop('DBSnapshotArn'), 'id': snapshot_id, 'name': snapshot_id, 'vpc_id': vpc_id}
        attributes = [
            'DBInstanceIdentifier',
            'SnapshotCreateTime',
            'Encrypted',
            'OptionGroupName'
        ]
        for attribute in attributes:
            snapshot[attribute] = dbs[attribute] if attribute in dbs else None
        api_client = api_clients[region]
        attributes = api_client.describe_db_snapshot_attributes(DBSnapshotIdentifier = snapshot_id)['DBSnapshotAttributesResult']
        snapshot['attributes'] = attributes['DBSnapshotAttributes'] if 'DBSnapshotAttributes' in attributes else {}
        # Save
        manage_dictionary(self.vpcs, vpc_id,RDSVPCConfig())
        self.vpcs[vpc_id].snapshots[snapshot_id] = snapshot


    def parse_parameter_group(self, global_params, region, parameter_group):
        parameter_group['arn'] = parameter_group.pop('DBParameterGroupArn')
        parameter_group['name'] = parameter_group.pop('DBParameterGroupName')
        api_client = api_clients[region]
        parameters = handle_truncated_response(api_client.describe_db_parameters, {'DBParameterGroupName': parameter_group['name']}, ['Parameters'])['Parameters']
        for parameter in parameters:
            param = {}
            param['value'] = parameter['ParameterValue'] if 'ParameterValue' in parameter else None
            param['source'] = parameter['Source']
            manage_dictionary(parameter_group, 'parameters', {})
            parameter_group['parameters'][parameter['ParameterName']] = param
        # Save
        (self).parameter_groups[parameter_group['name']] = parameter_group


    def parse_security_group(self, global_params, region, security_group):
        """
        Parse a single Redsfhit security group

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param security)_group:         Security group
        """
        vpc_id = security_group.pop('VpcId') if 'VpcId' in security_group else ec2_classic
        manage_dictionary(self.vpcs, vpc_id, RDSVPCConfig())
        security_group['arn'] = security_group.pop('DBSecurityGroupArn')
        security_group['name'] = security_group.pop('DBSecurityGroupName')
        # Save
        manage_dictionary(self.vpcs, vpc_id,RDSVPCConfig())
        self.vpcs[vpc_id].security_groups['name'] = security_group



########################################
# RDSConfig
########################################

class RDSConfig(RegionalServiceConfig):
    """
    RDS configuration for all AWS regions

    :cvar targets:                      Tuple with all RDS resource names that may be fetched
    :cvar config_class:                 Class to be used when initiating the service's configuration in a new region/VPC
    """
    targets = (
        ('instances', 'DBInstances', 'describe_db_instances', False),
        ('snapshots', 'DBSnapshots', 'describe_db_snapshots', False),
        ('parameter_groups', 'DBParameterGroups', 'describe_db_parameter_groups', False),
        ('security_groups', 'DBSecurityGroups', 'describe_db_security_groups', True),
        # TODO ('subnets', 'DBSubnetGroups', 'describe_db_subnet_group', False),
    )
    region_config_class = RDSRegionConfig



########################################
# RDSVPCConfig
########################################

class RDSVPCConfig(object):
    """
    RDS configuration for a single VPC

    :ivar instances:                     Dictionary of instances [id]
    :ivar security_groups:              Dictionary of security groups [name]
    """

    def __init__(self):
        self.instances = {}
        self.security_groups = {}
        self.snapshots = {}



def get_security_groups_info(rds_client, region_info):
    groups = rds_client.describe_db_security_groups()['DBSecurityGroups']
    manage_dictionary(region_info, 'vpcs', {})
    manage_dictionary(region_info['vpcs'], ec2_classic, {})
    manage_dictionary(region_info['vpcs'][ec2_classic], 'security_groups', {})
    manage_dictionary(region_info, 'security_groups_count', 0)
    region_info['security_groups_count'] += len(groups)
    for group in groups:
        region_info['vpcs'][ec2_classic]['security_groups'][group['DBSecurityGroupName']] = parse_security_group(group)

def parse_security_group(group):
    security_group = {}
    security_group['name'] = group['DBSecurityGroupName']
    security_group['description'] = group['DBSecurityGroupDescription']
    security_group['ec2_groups'] = {}
    for grant in group['EC2SecurityGroups']:
        if 'EC2SecurityGroupId' in grant:
            group_id = grant.pop('EC2SecurityGroupId')
        else:
            group_id = '%s-%s' % (grant['EC2SecurityGroupOwnerId'], grant['EC2SecurityGroupName'])
        security_group['ec2_groups'][group_id] = grant
    security_group['ip_ranges'] = {}
    for ip_range in group['IPRanges']:
        cidr = ip_range.pop('CIDRIP')
        security_group['ip_ranges'][cidr] = ip_range
    return security_group



# -*- coding: utf-8 -*-
"""
Single-service rule processing functions
"""

import copy
import re

from opinel.utils.conditions import pass_condition
from opinel.utils.console import printError
from opinel.utils.globals import manage_dictionary

from AWSScout2.rules import condition_operators
from AWSScout2.configs.browser import get_value_at



re_get_value_at = re.compile(r'_GET_VALUE_AT_\((.*?)\)')


def recurse(all_info, current_info, target_path, current_path, config, add_suffix = False):
    """

    :param all_info:
    :param current_info:
    :param target_path:
    :param current_path:
    :param config:
    :param add_suffix:
    :return:
    """
    results = []
    if len(target_path) == 0:
        # Dashboard: count the number of processed resources here
        manage_dictionary(config, 'checked_items', 0)
        config['checked_items'] = config['checked_items'] + 1
        # Test for conditions...
        if pass_conditions(all_info, current_path, copy.deepcopy(config['conditions'])):
            if add_suffix and 'id_suffix' in config:
                current_path.append(config['id_suffix'])
            results.append('.'.join(current_path))
        # Return the flagged items...
        config['flagged_items'] = len(results)
        return results
    target_path = copy.deepcopy(target_path)
    dbg_target_path = copy.deepcopy(target_path)
    current_path = copy.deepcopy(current_path)
    attribute = target_path.pop(0)
    if type(current_info) == dict:
        if attribute in current_info:
            split_path = copy.deepcopy(current_path)
            split_path.append(attribute)
            results = results + recurse(all_info, current_info[attribute], target_path, split_path, config, add_suffix)
        elif attribute == 'id':
            for key in current_info:
                split_target_path = copy.deepcopy(target_path)
                split_current_path = copy.deepcopy(current_path)
                split_current_path.append(key)
                split_current_info = current_info[key]
                results = results + recurse(all_info, split_current_info, split_target_path, split_current_path, config, add_suffix)
    # To handle lists properly, I would have to make sure the list is properly ordered and I can use the index to consistently access an object... Investigate (or do not use lists)
    elif type(current_info) == list:
        for index, split_current_info in enumerate(current_info):
            split_current_path = copy.deepcopy(current_path)
            split_current_path.append(str(index))
            results = results + recurse(all_info, split_current_info, copy.deepcopy(target_path), split_current_path, config, add_suffix)
    else:
        printError('Error: unhandled case, typeof(current_info) = %s' % type(current_info))
        printError('Path: %s' % current_path)
        printError('Object: %s' % str(current_info))
        printError('Entry target path: %s' % str(dbg_target_path))
        raise Exception
    return results


def pass_conditions(all_info, current_path, conditions):
    """
    Pass all conditions?

    :param all_info:
    :param current_path:
    :param conditions:
    :return:
    """
    result = False
    if len(conditions) == 0:
        return True
    condition_operator = conditions.pop(0)
    for condition in conditions:
      if condition[0] in condition_operators:
        res = pass_conditions(all_info, current_path, condition)
      else:
        # Conditions are formed as "path to value", "type of test", "value(s) for test"
        path_to_value, test_name, test_values = condition
        dynamic_path = re_get_value_at.findall(path_to_value)
        if dynamic_path:
            for dp in dynamic_path:
                dv = get_value_at(all_info, current_path, dp)
                path_to_value = path_to_value.replace('_GET_VALUE_AT_(%s)' % dp, dv)
        target_obj = get_value_at(all_info, current_path, path_to_value)
        if type(test_values) != list:
            dynamic_value = re_get_value_at.match(test_values)
            if dynamic_value:
                test_values = get_value_at(all_info, current_path, dynamic_value.groups()[0], True)
        res = pass_condition(target_obj, test_name, test_values)
      # Quick exit and + false
      if condition_operator == 'and' and not res:
          return False
      # Quick exit or + true
      if condition_operator == 'or' and res:
          return True
    # Still here ?
    # or -> false
    # and -> true
    if condition_operator == 'or':
        return False
    else:
        return True



# -*- coding: utf-8 -*-
"""
ELB-related classes and functions
"""

from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig
from AWSScout2.utils import ec2_classic, get_keys



########################################
# ELBRegionConfig
########################################

class ELBRegionConfig(RegionConfig):
    """
    ELB configuration for a single AWS region

    :ivar vpcs:                         Dictionary of VPCs [id]
    """

    def __init__(self):
        self.vpcs = {}


    def parse_elb(self, global_params, region, lb):
        """

        :param global_params:
        :param region:
        :param elb:
        :return:
        """
        elb = {}
        elb['name'] = lb.pop('LoadBalancerName')
        vpc_id = lb['VPCId'] if 'VPCId' in lb and lb['VPCId'] else ec2_classic
        manage_dictionary(self.vpcs, vpc_id, ELBVPCConfig())
        get_keys(lb, elb, ['DNSName', 'CreatedTime', 'AvailabilityZones', 'Subnets', 'Policies', 'Scheme'])
        elb['security_groups'] = []
        for sg in lb['SecurityGroups']:
            elb['security_groups'].append({'GroupId': sg})
        manage_dictionary(elb, 'listeners', {})
        for l in lb['ListenerDescriptions']:
            listener = l['Listener']
            manage_dictionary(listener, 'policies', {})
            for policy_name in l['PolicyNames']:
                manage_dictionary(listener['policies'], policy_name, {})
            elb['listeners'][l['Listener']['LoadBalancerPort']] = listener
        manage_dictionary(elb, 'instances', [])
        for i in lb['Instances']:
            elb['instances'].append(i['InstanceId'])
        self.vpcs[vpc_id].elbs[self.get_non_aws_id(elb['name'])] = elb



########################################
# ELBConfig
########################################

class ELBConfig(RegionalServiceConfig):
    """
    ELB configuration for all AWS regions

    :cvar targets:                      Tuple with all ELB resource names that may be fetched
    :cvar config_class:                 Class to be used when initiating the service's configuration in a new region/VPC
    """
    targets = (
        ('elbs', 'LoadBalancerDescriptions', 'describe_load_balancers', False),
    )
    region_config_class = ELBRegionConfig

    def finalize(self):
        pass



########################################
# ELBVPCConfig
########################################

class ELBVPCConfig(object):
    """
    ELB configuration for a single VPC

    :ivar flow_logs:                    Dictionary of flow logs [id]
    :ivar instances:                    Dictionary of instances [id]
    """

    def __init__(self, name = None):
        self.name = name
        self.elbs = {}




# -*- coding: utf-8 -*-

from opinel.utils.console import printError, printException, printInfo

from AWSScout2.utils import format_service_name
from AWSScout2.services.cloudformation import CloudFormationConfig
from AWSScout2.services.cloudtrail import CloudTrailConfig
from AWSScout2.services.cloudwatch import CloudWatchConfig
from AWSScout2.services.ec2 import EC2Config
from AWSScout2.services.elasticache import ElastiCacheConfig
from AWSScout2.services.elb import ELBConfig
from AWSScout2.services.elbv2 import ELBv2Config
from AWSScout2.services.iam import IAMConfig
from AWSScout2.services.rds import RDSConfig
from AWSScout2.services.redshift import RedshiftConfig
from AWSScout2.services.route53 import Route53Config, Route53DomainsConfig
from AWSScout2.services.s3 import S3Config
from AWSScout2.services.ses import SESConfig
from AWSScout2.services.sns import SNSConfig
from AWSScout2.services.sqs import SQSConfig
from AWSScout2.services.vpc import VPCConfig


class ServicesConfig(object):
    """
    Object that holds the necessary AWS configuration for all services in scope.
                                        
    :ivar cloudtrail:                   CloudTrail configuration
    :ivar cloudwatch:                   CloudWatch configuration: TODO
    :ivar ec2:                          EC2 configuration
    :ivar iam:                          IAM configuration
    :ivar rds:                          RDS configuration
    :ivar redshift:                     Redshift configuration
    :ivar s3:                           S3 configuration
    :ivar ses:                          SES configuration: TODO
    "ivar sns:                          SNS configuration
    :ivar sqs:                          SQS configuration
    """

    def __init__(self):
        self.cloudformation = CloudFormationConfig()
        self.cloudtrail = CloudTrailConfig()
        self.cloudwatch = CloudWatchConfig()
        self.ec2 = EC2Config()
        self.elasticache = ElastiCacheConfig()
        self.elb = ELBConfig()
        self.elbv2 = ELBv2Config()
        self.iam = IAMConfig()
        self.redshift = RedshiftConfig()
        self.rds = RDSConfig()
        self.route53 = Route53Config()
        self.route53domains = Route53DomainsConfig()
        self.s3 = S3Config()
        self.ses = SESConfig()
        self.sns = SNSConfig()
        self.sqs = SQSConfig()
        self.vpc = VPCConfig()


    def fetch(self, credentials, services = [], regions = [], partition_name = ''):
        """

        :param credentials:
        :param services:
        :param regions:
        :param partition_name:
        :return:
        """
        for service in vars(self):
            try:
                if services != [] and service not in services:
                    continue
                service_config = getattr(self, service)
                if 'fetch_all' in dir(service_config):
                    method_args = {}
                    method_args['credentials'] = credentials
                    if service != 'iam':
                        method_args['regions'] = regions
                        method_args['partition_name'] = partition_name
                    service_config.fetch_all(**method_args)
                    if hasattr(service_config, 'finalize'):
                        service_config.finalize()
            except Exception as e:
                printError('Error: could not fetch %s configuration.' % service)
                printException(e)

    def single_service_pass(self):
        pass

    def multi_service_pass(self):
        pass





def postprocessing(aws_config):
    for service in aws_config['services']:
        method_name = '%s_postprocessing' % service
        if method_name in globals():
            try:
                printInfo('Post-processing %s config...' % format_service_name(service))
                method = globals()[method_name]
                method(aws_config)
            except Exception as e:
                printException(e)
                pass






#!/usr/bin/env python
# -*- coding: utf-8 -*-

from opinel.utils.cli_parser import OpinelArgumentParser

from AWSScout2 import DEFAULT_REPORT_DIR



class SharedArgumentParser(OpinelArgumentParser):

    def __init__(self, default_args = None):
        super(SharedArgumentParser, self).__init__()
        self.add_argument('debug', default_args)
        self.add_argument('force', default_args)

    def add_argument(self, argument_name, default_args = None):
        if argument_name == 'services':
            self.parser.add_argument('--services',
                                    dest='services',
                                    default=[],
                                    nargs='+',
                                    help='Name of in-scope services.')
        elif argument_name == 'skip':
            self.parser.add_argument('--skip',
                                     dest='skipped_services',
                                     default=[],
                                     nargs='+',
                                     help='Name of out-of-scope services.')
        elif argument_name == 'timestamp':
            self.parser.add_argument('--timestamp',
                                     dest='timestamp',
                                     default=False,
                                     nargs='?',
                                     help='Timestamp added to the name of the report (default is current time in UTC).')
        elif argument_name == 'report-dir':
            self.parser.add_argument('--report-dir',
                                     dest='report_dir',
                                     default=DEFAULT_REPORT_DIR,
                                     help='Path of the Scout2 report.')
        elif argument_name == 'exceptions':
            self.parser.add_argument('--exceptions',
                                     dest='exceptions',
                                     default=[None],
                                     nargs='+',
                                     help='Exception file to use during analysis.')
        else:
            super(SharedArgumentParser, self).add_argument(argument_name, default_args)



class RulesArgumentParser(SharedArgumentParser):

    def __init__(self, default_args = None):
        super(RulesArgumentParser, self).__init__()
        self.parser.add_argument('--ruleset-name',
                                dest='ruleset_name',
                                default=None,
                                required=True,
                                help='Name of the ruleset to be generated.')
        self.parser.add_argument('--base-ruleset',
                                dest='base_ruleset',
                                default='default',
                                help='Ruleset to be used as the baseline.')
        self.parser.add_argument('--rules-dir',
                                 dest='rules_dir',
                                 default=[],
                                 nargs='+',
                                 help='Path to custom rule files.')
        self.parser.add_argument('--generator-dir',
                                 dest='generator_dir',
                                 default=DEFAULT_REPORT_DIR,
                                 help='Path of the Scout2 rules generator.')



class ListallArgumentParser(SharedArgumentParser):

    def __init__(self, default_args = None):
        super(ListallArgumentParser, self).__init__()
        self.add_argument('profile', default_args)
        self.add_argument('report-dir', default_args)
        self.add_argument('ip-ranges', default_args)
        self.add_argument('timestamp', default_args)
        self.add_argument('exceptions', default_args)
        self.parser.add_argument('--format',
                                dest='format',
                                default=['csv'],
                                nargs='+',
                                help='Listall output format.')
        self.parser.add_argument('--format-file',
                                dest='format_file',
                                default=None,
                                nargs='+',
                                help='Listall output format file')
        self.parser.add_argument('--config',
                                dest='config',
                                default=None,
                                help='Config file that sets the path and keys to be listed.')
        self.parser.add_argument('--config-args',
                                 dest='config_args',
                                 default=[],
                                 nargs='+',
                                 help='Arguments to be passed to the config file.')
        self.parser.add_argument('--path',
                                dest='path',
                                default=[],
                                nargs='+',
                                help='Path of the resources to list (e.g. iam.users.id or ec2.regions.id.vpcs.id)')
        self.parser.add_argument('--keys',
                                dest='keys',
                                default=[],
                                nargs='+',
                                help='Keys to be printed for the given object.')
        self.parser.add_argument('--keys-from-file',
                                dest='keys_file',
                                default=[],
                                nargs='+',
                                help='Keys to be printed for the given object (read values from file.')



class Scout2ArgumentParser(SharedArgumentParser):

    def __init__(self, default_args = None):
        super(Scout2ArgumentParser, self).__init__()
        self.add_argument('profile', default_args)
        self.add_argument('regions', default_args)
        self.add_argument('partition-name', default_args)
        self.add_argument('vpc', default_args)
        self.add_argument('ip-ranges', default_args)
        self.add_argument('ip-ranges-name-key', default_args)
        self.add_argument('mfa-serial')
        self.add_argument('mfa-code')
        self.add_argument('csv-credentials')
        self.add_argument('report-dir', default_args)
        self.add_argument('timestamp', default_args)
        self.add_argument('exceptions', default_args)
        self.add_argument('services', default_args)
        self.add_argument('skip', default_args)
        self.parser.add_argument('-l', '--local',
                                dest='fetch_local',
                                default=False,
                                action='store_true',
                                help='Use local data previously fetched and re-run the analyzis.')
        self.parser.add_argument('--resume',
                                dest='resume',
                                default=False,
                                action='store_true',
                                help='Complete a partial (throttled) run')
        self.parser.add_argument('--update',
                                dest='update',
                                default=False,
                                action='store_true',
                                help='Reload all the existing data and only overwrite data in scope for this run')
        self.parser.add_argument('--ruleset',
                                dest='ruleset',
                                default=None,
                                nargs='?',
                                help='Set of rules to be used during the analysis.')
        self.parser.add_argument('--no-browser',
                                 dest='no_browser',
                                 default=False,
                                 action='store_true',
                                 help='Do not automatically open the report in the browser.')

    def parse_args(self):
        args = self.parser.parse_args()
        # If local analysis, overwrite results
        if args.fetch_local:
            args.force_write = True
        return args



# -*- coding: utf-8 -*-
"""
Exceptions handling
"""



def process_exceptions(aws_config, exceptions_filename = None):
    """
    DDDD

    :param aws_config:
    :param exceptions_filename:
    :return:
    """

    # Load exceptions
    if not exceptions_filename:
        return
    with open(exceptions_filename, 'rt') as f:
        exceptions = json.load(f)

    # Process exceptions
        for service in exceptions['services']:
            for rule in exceptions['services'][service]['exceptions']:
                filtered_items = []
                for item in aws_config['services'][service]['violations'][rule]['items']:
                    if item not in exceptions['services'][service]['exceptions'][rule]:
                        filtered_items.append(item)
                aws_config['services'][service]['violations'][rule]['items'] = filtered_items
                aws_config['services'][service]['violations'][rule]['flagged_items'] = len(aws_config['services'][service]['violations'][rule]['items'])


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import webbrowser

try:
    from opinel.utils.console import configPrintException, printInfo
    from opinel.utils.globals import check_requirements
except Exception as e:
    print('Error: Scout2 depends on the opinel package. Install all the requirements with the following command:')
    print('  $ pip install -r requirements.txt')
    print(e)
    sys.exit(42)

from AWSScout2.cli_parser import RulesArgumentParser
from AWSScout2.rules.ruleset import Ruleset
from AWSScout2.output.html import RulesetGenerator


########################################
##### Main
########################################

def main():

    # Parse arguments
    parser = RulesArgumentParser()
    args = parser.parse_args()

    # Configure the debug level
    configPrintException(args.debug)

    # Check version of opinel
    if not check_requirements(os.path.realpath(__file__)):
        return 42

    # Load ruleset
    ruleset = Ruleset(filename = args.base_ruleset, name = args.ruleset_name, load_rules = False, rules_dir = args.rules_dir)

    # Generate the HTML generator
    ruleset_generator = RulesetGenerator(args.ruleset_name, args.generator_dir)
    ruleset_generator_path = ruleset_generator.save(ruleset, args.force_write, args.debug)

    # Open the HTML ruleset generator in a browser
    printInfo('Starting the HTML ruleset generator...')
    url = 'file://%s' % os.path.abspath(ruleset_generator_path)
    webbrowser.open(url, new=2)



# -*- coding: utf-8 -*-

import sys

from opinel.utils.console import printException

from AWSScout2 import __version__ as scout2_version



def postprocessing(aws_config, current_time, ruleset):
    """

    :param aws_config:
    :param current_time:
    :param ruleset:
    :return:
    """
    update_metadata(aws_config)
    update_last_run(aws_config, current_time, ruleset)


def update_last_run(aws_config, current_time, ruleset):
    last_run = {}
    last_run['time'] = current_time.strftime("%Y-%m-%d %H:%M:%S%z")
    last_run['cmd'] = ' '.join(sys.argv)
    last_run['version'] = scout2_version
    last_run['ruleset_name'] = ruleset.name
    last_run['ruleset_about'] = ruleset.ruleset['about'] if 'about' in ruleset.ruleset else ''
    last_run['summary'] = {}
    for service in aws_config['services']:
        last_run['summary'][service] = {'checked_items': 0, 'flagged_items': 0, 'max_level': 'warning', 'rules_count': 0}
        if aws_config['services'][service] == None:
            # Not supported yet
            continue
        elif 'findings' in aws_config['services'][service]:
            for finding in aws_config['services'][service]['findings']:
                last_run['summary'][service]['rules_count'] += 1
                last_run['summary'][service]['checked_items'] += aws_config['services'][service]['findings'][finding]['checked_items']
                last_run['summary'][service]['flagged_items'] += aws_config['services'][service]['findings'][finding]['flagged_items']
                if last_run['summary'][service]['max_level'] != 'danger':
                    last_run['summary'][service]['max_level'] = aws_config['services'][service]['findings'][finding]['level']
    aws_config['last_run'] = last_run


def update_metadata(aws_config):
        service_map = {}
        for service_group in aws_config['metadata']:
            for service in aws_config['metadata'][service_group]:
                if service not in aws_config['service_list']:
                    continue
                if 'resources' not in aws_config['metadata'][service_group][service]:
                    continue
                service_map[service] = service_group
                for resource in aws_config['metadata'][service_group][service]['resources']:
                    # full_path = path if needed
                    if not 'full_path' in aws_config['metadata'][service_group][service]['resources'][resource]:
                        aws_config['metadata'][service_group][service]['resources'][resource]['full_path'] = aws_config['metadata'][service_group][service]['resources'][resource]['path']
                    # Script is the full path minus "id" (TODO: change that)
                    if not 'script' in aws_config['metadata'][service_group][service]['resources'][resource]:
                        aws_config['metadata'][service_group][service]['resources'][resource]['script'] = '.'.join([x for x in aws_config['metadata'][service_group][service]['resources'][resource]['full_path'].split('.') if x != 'id'])
                    # Update counts
                    count = '%s_count' % resource
                    service_config = aws_config['services'][service]
                    if service_config and resource != 'regions':
                      if 'regions' in service_config.keys(): # hasattr(service_config, 'regions'):
                        aws_config['metadata'][service_group][service]['resources'][resource]['count'] = 0
                        for region in service_config['regions']:
                            if count in service_config['regions'][region].keys():
                                aws_config['metadata'][service_group][service]['resources'][resource]['count'] += service_config['regions'][region][count]
                      else:
                          try:
                            aws_config['metadata'][service_group][service]['resources'][resource]['count'] = service_config[count]
                          except Exception as e:

                              printException(e)



#!/usr/bin/env python

# distutils/setuptools install script for Scout2
import os
from setuptools import setup, find_packages

# Package info
NAME = 'AWSScout2'
ROOT = os.path.dirname(__file__)
VERSION = __import__(NAME).__version__

# Requirements
requirements = []
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'AWSScout2', 'data', 'requirements.txt')) as f:
    for r in f.readlines():
        requirements.append(r.strip())

# Setup
setup(
    name=NAME,
    version=VERSION,
    description='Scout2, TODO',
    long_description=open('README.rst').read(),
    author='l01cd3v',
    author_email='l01cd3v@gmail.com',
    url='https://github.com/nccgroup/Scout2',
    entry_points={
        'console_scripts': [
            'Scout2 = AWSScout2.__main__:main',
            'Scout2RulesGenerator = AWSScout2.__rules_generator__:main',
            'Scout2Listall = AWSScout2.__listall__:main'
        ]
    },
    packages=[
        'AWSScout2', 'AWSScout2.configs', 'AWSScout2.output', 'AWSScout2.rules', 'AWSScout2.services'
    ],
    package_data={
        'AWSScout2': [
            'data/requirements.txt'
        ],
        'AWSScout2.configs': [
            'data/*.json'
        ],
        'AWSScout2.output': [
            'data/html/*.html',
            'data/html/partials/*.html',
            'data/html/summaries/*.html',
            'data/includes.zip',
            'data/inc-scout2/*.js',
            'data/inc-scout2/*.css'
        ],
        'AWSScout2.rules': [
            'data/*.html',
            'data/filters/*.json',
            'data/findings/*.json',
            'data/rulesets/*.json'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license='GNU General Public License v2 (GPLv2)',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],
)



# -*- coding: utf-8 -*-

from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients
from AWSScout2.utils import ec2_classic



########################################
# ElastiCacheRegionConfig
########################################

class ElastiCacheRegionConfig(RegionConfig):
    """
    ElastiCache configuration for a single AWS region

    :ivar vpcs:                         Dictionary of VPCs [id]
    :ivar clusters_count:               Number of clusters in the region
    :ivar security_groups:              Dictionary of security groups [name]
    :ivar security_groups_count:        Number of security groups in the region
    """

    def __init__(self):
        self.vpcs = {}
        self.clusters_count = 0
        self.parameter_groups = {}
        self.parameter_groups_count = 0
        self.security_groups = {}
        self.security_groups_count = 0


    def parse_cluster(self, global_params, region, cluster):
        """
        Parse a single ElastiCache cluster

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param cluster:                 ElastiCache cluster
        """
        cluster_name = cluster.pop('CacheClusterId')
        cluster['name'] = cluster_name
        # Must fetch info about the subnet group to retrieve the VPC ID...
        if 'CacheSubnetGroupName' in cluster:
            subnet_group = api_clients[region].describe_cache_subnet_groups(CacheSubnetGroupName = cluster['CacheSubnetGroupName'])['CacheSubnetGroups'][0]
            vpc_id = subnet_group['VpcId']
        else:
            vpc_id = ec2_classic
            subnet_group = None
        manage_dictionary(self.vpcs, vpc_id, ElastiCacheVPCConfig())
        self.vpcs[vpc_id].clusters[cluster_name] = cluster
        if subnet_group:
            self.vpcs[vpc_id].subnet_groups[subnet_group['CacheSubnetGroupName']] = subnet_group


    def parse_security_group(self, global_params, region, security_group):
        """
        Parse a single ElastiCache security group

        :param global_params:
        :param region:
        :param security_group:
        :return:
        """
        security_group['name'] = security_group.pop('CacheSecurityGroupName')
        self.security_groups[security_group['name']] = security_group



########################################
# ElastiCacheConfig
########################################

class ElastiCacheConfig(RegionalServiceConfig):
    """
    ElastiCache configuration for all AWS regions

    :cvar targets:                      Tuple with all ElastiCache resource names that may be fetched
    :cvar config_class:                 Class to be used when initiating the service's configuration in a new region/VPC
    """
    targets = (
        ('clusters', 'CacheClusters', 'describe_cache_clusters', False),
        ('security_groups', 'CacheSecurityGroups', 'describe_cache_security_groups', True)      # Do not throw errors as this is disabled in newer regions
        #('parameter_groups', 'DBParameterGroups', 'describe_db_parameter_groups', False),
    )
    region_config_class = ElastiCacheRegionConfig



########################################
# ElastiCacheVPCConfig
########################################

class ElastiCacheVPCConfig(object):
    """
    ElastiCache configuration for a single VPC

    :ivar clusters:                     Dictionary of clusters [name]
    :ivar subnet_groups:                Dictionary of subnet groups [name]
    """

    def __init__(self):
        self.clusters = {}
        self.subnet_groups = {}



# -*- coding: utf-8 -*-

import json

from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients



########################################
# SESRegionConfig
########################################

class SESRegionConfig(RegionConfig):
    """
    SES configuration for a single AWS region

    :ivar identities:                       Dictionary of identities [name]
    :ivar identities_count:                 Number of identities in the region
    """

    def __init__(self):
        self.identities = {}
        self.identities_count = 0


    def parse_identitie(self, global_params, region, identity_name):
        """
        Parse a single identity and fetch additional attributes

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        """
        identity = {'name': identity_name, 'policies': {}}
        policy_names = api_clients[region].list_identity_policies(Identity = identity_name)['PolicyNames']
        if len(policy_names):
            policies = api_clients[region].get_identity_policies(Identity = identity_name, PolicyNames = policy_names)['Policies']
            for policy_name in policies:
                identity['policies'][policy_name] = json.loads(policies[policy_name])
        self.identities[self.get_non_aws_id(identity_name)] = identity



########################################
# SESConfig
########################################

class SESConfig(RegionalServiceConfig):
    """
    SES configuration for all AWS regions

    :cvar targets:                      Tuple with all SES resource names that may be fetched
    :cvar region_config_class:          Class to be used when initiating the service's configuration in a new region
    """
    targets = (
        ('identities', 'Identities', 'list_identities', False),
    )
    region_config_class = SESRegionConfig


# -*- coding: utf-8 -*-

import copy

from opinel.utils.console import printDebug, printError, printException, printInfo
from opinel.utils.globals import manage_dictionary

from AWSScout2.configs.browser import combine_paths, get_object_at, get_value_at
from AWSScout2.utils import ec2_classic
from AWSScout2.services.vpc import put_cidr_name



def preprocessing(aws_config, ip_ranges = [], ip_ranges_name_key = None):
    """
    Tweak the AWS config to match cross-service resources and clean any fetching artifacts

    :param aws_config:
    :return:
    """
    set_aws_account_id(aws_config)
    sort_vpc_flow_logs(aws_config['services']['vpc'])
    sort_elbs(aws_config)
    list_ec2_network_attack_surface(aws_config['services']['ec2'])
    add_security_group_name_to_ec2_grants(aws_config['services']['ec2'], aws_config['aws_account_id'])
    process_cloudtrail_trails(aws_config['services']['cloudtrail'])
    process_network_acls(aws_config['services']['vpc'])
    match_network_acls_and_subnets(aws_config['services']['vpc'])
    match_instances_and_roles(aws_config)
    match_roles_and_cloudformation_stacks(aws_config)
    match_roles_and_vpc_flowlogs(aws_config)
    match_iam_policies_and_buckets(aws_config)
    match_security_groups_and_resources(aws_config)
    add_cidr_display_name(aws_config, ip_ranges, ip_ranges_name_key)
    merge_route53_and_route53domains(aws_config)



def add_cidr_display_name(aws_config, ip_ranges, ip_ranges_name_key):
    if len(ip_ranges):
        callback_args = {'ip_ranges': ip_ranges, 'ip_ranges_name_key': ip_ranges_name_key}
        go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports'], ['services', 'ec2'], put_cidr_name, callback_args)


def add_security_group_name_to_ec2_grants(ec2_config, aws_account_id):
    """
    Github issue #24: display the security group names in the list of grants (added here to have ligher JS code)

    :param ec2_config:
    :param aws_account_id:
    :return:
    """
    go_to_and_do(ec2_config, None, ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports', 'security_groups'], [], add_security_group_name_to_ec2_grants_callback, {'AWSAccountId': aws_account_id})


def add_security_group_name_to_ec2_grants_callback(ec2_config, current_config, path, current_path, ec2_grant, callback_args):
    sg_id = ec2_grant['GroupId']
    if sg_id in current_path:
        target = current_path[:(current_path.index(sg_id) + 1)]
        ec2_grant['GroupName'] = get_value_at(ec2_config, target, 'name')
    elif ec2_grant['UserId'] == callback_args['AWSAccountId']:
        if 'VpcId' in ec2_grant:
            target = current_path[:(current_path.index('vpcs') + 1)]
            target.append(ec2_grant['VpcId'])
            target.append('security_groups')
            target.append(sg_id)
        else:
            target = current_path[:(current_path.index('security_groups') + 1)]
            target.append(sg_id)
        ec2_grant['GroupName'] = get_value_at(ec2_config, target, 'name')


def process_cloudtrail_trails(cloudtrail_config):
    printInfo('Processing CloudTrail config...')
    global_events_logging = []
    for region in cloudtrail_config['regions']:
        for trail_id in cloudtrail_config['regions'][region]['trails']:
            trail = cloudtrail_config['regions'][region]['trails'][trail_id]
            if 'HomeRegion' in trail and trail['HomeRegion'] != region:
                # Part of a multi-region trail, skip until we find the whole object
                continue
            if trail['IncludeGlobalServiceEvents'] == True and trail['IsLogging'] == True:
                global_events_logging.append((region, trail_id,))
    cloudtrail_config['IncludeGlobalServiceEvents'] = False if (len(global_events_logging) == 0) else True
    cloudtrail_config['DuplicatedGlobalServiceEvents'] = True if (len(global_events_logging) > 1) else False


def process_network_acls(vpc_config):
    printInfo('Analyzing VPC Network ACLs...')
    go_to_and_do(vpc_config, None, ['regions', 'vpcs', 'network_acls'], [], process_network_acls_callback, {})


def process_network_acls_callback(vpc_config, current_config, path, current_path, privateip_id, callback_args):
    # Check if the network ACL allows all traffic from all IP addresses
    process_network_acls_check_for_allow_all(current_config, 'ingress')
    process_network_acls_check_for_allow_all(current_config, 'egress')
    # Check if the network ACL only has the default rules
    process_network_acls_check_for_default(current_config, 'ingress')
    process_network_acls_check_for_default(current_config, 'egress')


def process_network_acls_check_for_allow_all(network_acl, direction):
    network_acl['allow_all_%s_traffic' % direction] = False
    for rule in network_acl['rules'][direction]:
        if rule['RuleAction'] == 'deny':
            # If a deny rule appears before an allow all, do not raise the flag
            break
        if (rule['CidrBlock'] == '0.0.0.0/0') and (rule['RuleAction'] == 'allow') and (rule['port_range'] == '1-65535') and (rule['protocol'] == 'ALL'):
            network_acl['allow_all_%s_traffic' % direction] = True
            break


def process_network_acls_check_for_default(network_acl, direction):
    if len(network_acl['rules'][direction]) != 2 and network_acl['allow_all_%s_traffic' % direction] != False:
        network_acl['use_default_%s_rules' % direction] = False
    else:
        # Assume it's the default rules because 2 rules and there's an allow all
        network_acl['use_default_%s_rules' % direction] = True


def list_ec2_network_attack_surface(ec2_config):
    go_to_and_do(ec2_config, None, ['regions', 'vpcs', 'instances', 'network_interfaces', 'PrivateIpAddresses'], [], list_ec2_network_attack_surface_callback, {})


def list_ec2_network_attack_surface_callback(ec2_config, current_config, path, current_path, privateip_id, callback_args):
    if 'Association' in current_config and current_config['Association']:
        public_ip = current_config['Association']['PublicIp']
        manage_dictionary(ec2_config, 'attack_surface', {})
        manage_dictionary(ec2_config['attack_surface'], public_ip, {'protocols': {}})
        for sg_info in current_config['Groups']:
            sg_id = sg_info['GroupId']
            sg_path = copy.deepcopy(current_path[0:4])
            sg_path.append('security_groups')
            sg_path.append(sg_id)
            sg_path.append('rules')
            sg_path.append('ingress')
            ingress_rules = get_object_at(ec2_config, sg_path)
            public_ip_grants = {}
            for p in ingress_rules['protocols']:
                for port in ingress_rules['protocols'][p]['ports']:
                    if 'cidrs' in ingress_rules['protocols'][p]['ports'][port]:
                        manage_dictionary(ec2_config['attack_surface'][public_ip]['protocols'], p, {'ports': {}})
                        manage_dictionary(ec2_config['attack_surface'][public_ip]['protocols'][p]['ports'], port, {'cidrs': []})
                        ec2_config['attack_surface'][public_ip]['protocols'][p]['ports'][port]['cidrs'] += ingress_rules['protocols'][p]['ports'][port]['cidrs']


def match_iam_policies_and_buckets(aws_config):
    s3_info = aws_config['services']['s3']
    iam_info = aws_config['services']['iam']
    if 'Action' in iam_info['permissions']:
        for action in (x for x in iam_info['permissions']['Action'] if ((x.startswith('s3:') and x != 's3:ListAllMyBuckets') or (x == '*'))):
            for iam_entity in iam_info['permissions']['Action'][action]:
                if 'Allow' in iam_info['permissions']['Action'][action][iam_entity]:
                    for allowed_iam_entity in iam_info['permissions']['Action'][action][iam_entity]['Allow']:
                        # For resource statements, we can easily rely on the existing permissions structure
                        if 'Resource' in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]:
                            for full_path in (x for x in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['Resource'] if x.startswith('arn:aws:s3:') or x == '*'):
                                parts = full_path.split('/')
                                bucket_name = parts[0].split(':')[-1]
                                __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['Resource'][full_path])
                        # For notresource statements, we must fetch the policy document to determine which buckets are not protected
                        if 'NotResource' in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]:
                            for full_path in (x for x in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'] if x.startswith('arn:aws:s3:') or x == '*'):
                                for policy_type in ['InlinePolicies', 'ManagedPolicies']:
                                    if policy_type in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path]:
                                        for policy in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path][policy_type]:
                                            __update_bucket_permissions(s3_info, iam_info, action, iam_entity, allowed_iam_entity, full_path, policy_type, policy)

def __update_bucket_permissions(s3_info, iam_info, action, iam_entity, allowed_iam_entity, full_path, policy_type, policy_name):
    allowed_buckets = []
    # By default, all buckets are allowed
    for bucket_name in s3_info['buckets']:
        allowed_buckets.append(bucket_name)
    if policy_type == 'InlinePolicies':
        policy = iam_info[iam_entity.title()][allowed_iam_entity]['Policies'][policy_name]['PolicyDocument']
    elif policy_type == 'ManagedPolicies':
        policy = iam_info['ManagedPolicies'][policy_name]['PolicyDocument']
    else:
        printError('Error, found unknown policy type.')
    for statement in policy['Statement']:
        for target_path in statement['NotResource']:
            parts = target_path.split('/')
            bucket_name = parts[0].split(':')[-1]
            path = '/' + '/'.join(parts[1:]) if len(parts) > 1 else '/'
            if (path == '/' or path == '/*') and (bucket_name in allowed_buckets):
                # Remove bucket from list
                allowed_buckets.remove(bucket_name)
            elif bucket_name == '*':
                allowed_buckets = []
    policy_info = {}
    policy_info[policy_type] = {}
    policy_info[policy_type][policy_name] = iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path][policy_type][policy_name]
    for bucket_name in allowed_buckets:
        __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info)

def __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info):
    if bucket_name != '*' and bucket_name in s3_info['buckets']:
        bucket = s3_info['buckets'][bucket_name]
        manage_dictionary(bucket, iam_entity, {})
        manage_dictionary(bucket, iam_entity + '_count', 0)
        if not allowed_iam_entity in bucket[iam_entity]:
            bucket[iam_entity][allowed_iam_entity] = {}
            bucket[iam_entity + '_count'] = bucket[iam_entity + '_count'] + 1

        if 'inline_policies' in policy_info:
            manage_dictionary(bucket[iam_entity][allowed_iam_entity], 'inline_policies', {})
            bucket[iam_entity][allowed_iam_entity]['inline_policies'].update(policy_info['inline_policies'])
        if 'policies' in policy_info:
            manage_dictionary(bucket[iam_entity][allowed_iam_entity], 'policies', {})
            bucket[iam_entity][allowed_iam_entity]['policies'].update(policy_info['policies'])
    elif bucket_name == '*':
        for bucket in s3_info['buckets']:
            __update_iam_permissions(s3_info, bucket, iam_entity, allowed_iam_entity, policy_info)
        pass
    else:
        # Could be an error or cross-account access, ignore...
        pass


def match_network_acls_and_subnets(vpc_config):
    printInfo('Matching VPC network ACLs and subnets...')
    go_to_and_do(vpc_config, vpc_config, ['regions', 'vpcs', 'network_acls'], [], match_network_acls_and_subnets_callback, {})


def match_network_acls_and_subnets_callback(vpc_config, current_config, path, current_path, acl_id, callback_args):
    for association in current_config['Associations']:
        subnet_path = current_path[:-1] + ['subnets', association['SubnetId']]
        subnet = get_object_at(vpc_config, subnet_path)
        subnet['network_acl'] = acl_id


def match_instances_and_roles(aws_config):
    """
    Foobar

    :param aws_config:
    """
    printInfo('Matching EC2 instances and IAM roles...')
    ec2_config = aws_config['services']['ec2']
    iam_config = aws_config['services']['iam']
    role_instances = {}
    for r in ec2_config['regions']:
        for v in ec2_config['regions'][r]['vpcs']:
            if 'instances' in ec2_config['regions'][r]['vpcs'][v]:
                for i in ec2_config['regions'][r]['vpcs'][v]['instances']:
                    instance_profile_id = ec2_config['regions'][r]['vpcs'][v]['instances'][i]['iam_instance_profile']['id'] if 'iam_instance_profile' in ec2_config['regions'][r]['vpcs'][v]['instances'][i] else None
                    if instance_profile_id:
                        manage_dictionary(role_instances, instance_profile_id, [])
                        role_instances[instance_profile_id].append(i)
    for role_id in iam_config['roles']:
        iam_config['roles'][role_id]['instances_count'] = 0
        for instance_profile_id in iam_config['roles'][role_id]['instance_profiles']:
            if instance_profile_id in role_instances:
                iam_config['roles'][role_id]['instance_profiles'][instance_profile_id]['instances'] = role_instances[instance_profile_id]
                iam_config['roles'][role_id]['instances_count'] += len(role_instances[instance_profile_id])


def match_roles_and_cloudformation_stacks(aws_config):
    go_to_and_do(aws_config, aws_config['services']['cloudformation'], ['regions', 'stacks'], [], match_roles_and_cloudformation_stacks_callback, {})


def match_roles_and_cloudformation_stacks_callback(aws_config, current_config, path, current_path, stack_id, callback_args):
    if 'RoleARN' not in current_config:
        return
    role_arn = current_config.pop('RoleARN')
    current_config['iam_role'] = __get_role_info(aws_config, 'arn', role_arn)


def match_roles_and_vpc_flowlogs(aws_config):
    go_to_and_do(aws_config, aws_config['services']['vpc'], ['regions', 'flow_logs'], [], match_roles_and_vpc_flowlogs_callback, {})


def match_roles_and_vpc_flowlogs_callback(aws_config, current_config, path, current_path, flowlog_id, callback_args):
    if 'DeliverLogsPermissionArn' not in current_config:
        return
    delivery_role_arn = current_config.pop('DeliverLogsPermissionArn')
    current_config['delivery_role'] = __get_role_info(aws_config, 'arn', delivery_role_arn)


def __get_role_info(aws_config, attribute_name, attribute_value):
    iam_role_info = {'name': None, 'id': None}
    for role_id in aws_config['services']['iam']['roles']:
        if aws_config['services']['iam']['roles'][role_id][attribute_name] == attribute_value:
            iam_role_info['name'] = aws_config['services']['iam']['roles'][role_id]['name']
            iam_role_info['id'] = role_id
            break
    return iam_role_info


def match_security_groups_and_resources(aws_config):
    # EC2 instances
    callback_args = {'status_path': [ '..', '..', 'State', 'Name' ], 'resource_id_path': ['..'], 'sg_list_attribute_name': 'Groups', 'sg_id_attribute_name': 'GroupId'}
    go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'instances', 'network_interfaces'], ['services', 'ec2'], match_security_groups_and_resources_callback, callback_args)
    # ELBs
    callback_args = {'status_path': ['Scheme'], 'sg_list_attribute_name': 'security_groups', 'sg_id_attribute_name': 'GroupId'}
    go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'elbs'], ['services', 'ec2'], match_security_groups_and_resources_callback, callback_args)
    # Redshift clusters
    callback_args = {'status_path': ['ClusterStatus'], 'sg_list_attribute_name': 'VpcSecurityGroups', 'sg_id_attribute_name': 'VpcSecurityGroupId'}
    go_to_and_do(aws_config, aws_config['services']['redshift'], ['regions', 'vpcs', 'clusters'], ['services', 'redshift'], match_security_groups_and_resources_callback, callback_args)
    # RDS instances
    callback_args = {'status_path': ['DBInstanceStatus'], 'sg_list_attribute_name': 'VpcSecurityGroups', 'sg_id_attribute_name': 'VpcSecurityGroupId'}
    go_to_and_do(aws_config, aws_config['services']['rds'], ['regions', 'vpcs', 'instances'], ['services', 'rds'], match_security_groups_and_resources_callback, callback_args)
    # ElastiCache clusters
    callback_args = {'status_path': ['CacheClusterStatus'], 'sg_list_attribute_name': 'SecurityGroups', 'sg_id_attribute_name': 'SecurityGroupId'}
    go_to_and_do(aws_config, aws_config['services']['elasticache'], ['regions', 'vpcs', 'clusters'], ['services', 'elasticache'], match_security_groups_and_resources_callback, callback_args)


def match_security_groups_and_resources_callback(aws_config, current_config, path, current_path, resource_id, callback_args):
    service = current_path[1]
    original_resource_path = combine_paths(copy.deepcopy(current_path), [ resource_id ])
    resource = get_object_at(aws_config, original_resource_path)
    if not 'resource_id_path' in callback_args:
        resource_type = current_path[-1]
        resource_path = copy.deepcopy(current_path)
        resource_path.append(resource_id)
    else:
        resource_path = combine_paths(copy.deepcopy(current_path), callback_args['resource_id_path'])
        resource_id = resource_path[-1]
        resource_type = resource_path[-2]
    #print('Resource path: %s' % resource_path)
    #print('Resource type: %s' % resource_type)
    #print('Resource id: %s' % resource_id)
    if 'status_path' in callback_args:
        status_path = combine_paths(copy.deepcopy(original_resource_path), callback_args['status_path'])
        #print('Status path: %s' % status_path)
        resource_status = get_object_at(aws_config, status_path)
    else:
        resource_status = None
    sg_base_path = copy.deepcopy(current_path[0:6])
    sg_base_path[1] = 'ec2'
    sg_base_path.append('security_groups')
    # Issue 89 & 91 : can instances have no security group?
    try:
        for resource_sg in resource[callback_args['sg_list_attribute_name']]:
            sg_id = resource_sg[callback_args['sg_id_attribute_name']]
            sg_path = copy.deepcopy(sg_base_path)
            sg_path.append(sg_id)
            sg = get_object_at(aws_config, sg_path)
            # Add usage information
            manage_dictionary(sg, 'used_by', {})
            manage_dictionary(sg['used_by'], service, {})
            manage_dictionary(sg['used_by'][service], 'resource_type', {})
            manage_dictionary(sg['used_by'][service]['resource_type'], resource_type, {} if resource_status else [])
            if resource_status:
                manage_dictionary(sg['used_by'][service]['resource_type'][resource_type], resource_status, [])
                if not resource_id in sg['used_by'][service]['resource_type'][resource_type][resource_status]:
                    sg['used_by'][service]['resource_type'][resource_type][resource_status].append(resource_id)
            else:
                sg['used_by'][service]['resource_type'][resource_type].append(resource_id)
    except Exception as e:
        region = current_path[3]
        vpc_id = current_path[5]
        if vpc_id == ec2_classic and resource_type == 'elbs':
            pass
        else:
            printError('Failed to parse %s in %s in %s' % (resource_type, vpc_id, region))
            printException(e)


def merge_route53_and_route53domains(aws_config):
    if 'route53domains' not in aws_config['services']:
        return
    aws_config['services']['route53'].update(aws_config['services']['route53domains'])
    aws_config['services'].pop('route53domains')


def set_aws_account_id(aws_config):
    for rt in ['groups', 'policies', 'roles', 'users']:
        for r in aws_config['services']['iam'][rt]:
            if 'aws_account_id' not in aws_config or aws_config['aws_account_id'] == None:
                if 'arn' in aws_config['services']['iam'][rt][r]:
                    aws_config['aws_account_id'] = aws_config['services']['iam'][rt][r]['arn'].split(':')[4]
        if 'aws_account_id' in aws_config and aws_config['aws_account_id'] != None:
            break
    if 'aws_account_id' not in aws_config:
        aws_config['aws_account_id'] = None


def sort_elbs(aws_config):
    """
    ELB and ELBv2 are different services, but for consistency w/ the console move them to the EC2 config

    :param aws_config:
    :return:
    """
    if 'elb' in aws_config['services']:
        go_to_and_do(aws_config, aws_config['services']['elb'], ['regions', 'vpcs', 'elbs'], [], sort_elbs_callback, {})
        aws_config['services'].pop('elb')
    if 'elbv2' in aws_config['services']:
        go_to_and_do(aws_config, aws_config['services']['elbv2'], ['regions', 'vpcs', 'elbs'], [], sort_elbs_callback, {})
        aws_config['services'].pop('elbv2')


def sort_elbs_callback(aws_config, current_config, path, current_path, elb_id, callback_args):
    vpc_config = get_object_at(aws_config, ['services', 'ec2'] + current_path[:-1])
    manage_dictionary(vpc_config, 'elbs', {})
    vpc_config['elbs'][elb_id] = current_config


def sort_vpc_flow_logs(vpc_config):
    go_to_and_do(vpc_config, None, ['regions', 'flow_logs'], [], sort_vpc_flow_logs_callback, {})


def sort_vpc_flow_logs_callback(vpc_config, current_config, path, current_path, flow_log_id, callback_args):
    attached_resource = current_config['ResourceId']
    if attached_resource.startswith('vpc-'):
        vpc_path = combine_paths(current_path[0:2], ['vpcs', attached_resource])
        try:
            attached_vpc = get_object_at(vpc_config, vpc_path)
        except Exception as e:
            printDebug('It appears that the flow log %s is attached to a resource that was previously deleted (%s).' % (flow_log_id, attached_resource))
            return
        manage_dictionary(attached_vpc, 'flow_logs', [])
        if flow_log_id not in attached_vpc['flow_logs']:
            attached_vpc['flow_logs'].append(flow_log_id)
        for subnet_id in attached_vpc['subnets']:
            manage_dictionary(attached_vpc['subnets'][subnet_id], 'flow_logs', [])
            if flow_log_id not in attached_vpc['subnets'][subnet_id]['flow_logs']:
                attached_vpc['subnets'][subnet_id]['flow_logs'].append(flow_log_id)
    elif attached_resource.startswith('subnet-'):
        all_vpcs = get_object_at(vpc_config, combine_paths(current_path[0:2], ['vpcs']))
        for vpc in all_vpcs:
            if attached_resource in all_vpcs[vpc]['subnets']:
                manage_dictionary(all_vpcs[vpc]['subnets'][attached_resource], 'flow_logs', [])
                if flow_log_id not in all_vpcs[vpc]['subnets'][attached_resource]['flow_logs']:
                    all_vpcs[vpc]['subnets'][attached_resource]['flow_logs'].append(flow_log_id)
                break
    else:
        printError('Resource %s attached to flow logs is not handled' % attached_resource)


def go_to_and_do(aws_config, current_config, path, current_path, callback, callback_args = None):
    """
    Recursively go to a target and execute a callback

    :param aws_config:                  A
    :param current_config:
    :param path:
    :param current_path:
    :param callback:
    :param callback_args:
    :return:
    """
    try:
        key = path.pop(0)
        if not current_config:
            current_config = aws_config
        if not current_path:
            current_path = []
        keys = key.split('.')
        if len(keys) > 1:
            while True:
                key = keys.pop(0)
                if not len(keys):
                    break
                current_path.append(key)
                current_config = current_config[key]
        if key in current_config:
            current_path.append(key)
            for (i, value) in enumerate(list(current_config[key])):
                if len(path) == 0:
                    if type(current_config[key] == dict) and type(value) != dict and type(value) != list:
                        callback(aws_config, current_config[key][value], path, current_path, value, callback_args)
                    else:
                        callback(aws_config, current_config, path, current_path, value, callback_args)
                else:
                    tmp = copy.deepcopy(current_path)
                    try:
                        tmp.append(value)
                        go_to_and_do(aws_config, current_config[key][value], copy.deepcopy(path), tmp, callback, callback_args)
                    except:
                        tmp.pop()
                        tmp.append(i)
                        go_to_and_do(aws_config, current_config[key][i], copy.deepcopy(path), tmp, callback, callback_args)

    except Exception as e:
        printException(e)
        if i:
            printInfo('Index: %s' % str(i))
        printInfo('Path: %s' % str(current_path))
        printInfo('Key = %s' % str(key))
        printInfo('Value = %s' % str(value))
        printInfo('Path = %s' % path)
